# -*- coding: utf-8 -*-


# Built-in
import warnings


# Common
import numpy as np
from matplotlib.tri import Triangulation as mplTri
import datastock as ds


_ELEMENTS = 'knots'


# #############################################################################
# #############################################################################
#               DEPRECATED ?
# #############################################################################


# def _get_RZ(arr, name=None, shapeRZ=None):
    # if arr.ndim == 1:
        # if np.any(np.diff(arr) <= 0.):
            # msg = "Non-increasing {}".format(name)
            # raise Exception(msg)
    # else:
        # lc = [np.all(np.diff(arr[0, :])) > 0.,
              # np.all(np.diff(arr[:, 0])) > 0.]
        # if np.sum(lc) != 1:
            # msg = "Impossible to know {} dimension!".format(name)
            # raise Exception(msg)
        # if lc[0]:
            # arr = arr[0, :]
            # if shapeRZ[1] is None:
                # shapeRZ[1] = name
            # if shapeRZ[1] != name:
                # msg = "Inconsistent shapeRZ"
                # raise Exception(msg)
        # else:
            # arr = arr[:, 0]
            # if shapeRZ[0] is None:
                # shapeRZ[0] = name
            # if shapeRZ[0] != name:
                # msg = "Inconsistent shapeRZ"
                # raise Exception(msg)
    # return arr, shapeRZ


# def _duplicates(arr, arru, nn, name=None, msg=None):
    # msg += (
        # "  Duplicate {}: {}\n".format(name, nn - arru.shape[0])
        # + "\t- {}.shape: {}\n".format(name, arr.shape)
        # + "\t- unique shape: {}".format(arru.shape)
    # )
    # return msg


# def _check_trimesh_conformity(nodes, faces, key=None):
    # nnodes = nodes.shape[0]
    # nfaces = faces.shape[0]

    # # Test for duplicates
    # nodesu = np.unique(nodes, axis=0)
    # facesu = np.unique(faces, axis=0)
    # lc = [nodesu.shape[0] != nnodes,
          # facesu.shape[0] != nfaces]
    # if any(lc):
        # msg = "Non-valid mesh ddata[{0}]: \n".format(key)
        # if lc[0]:
            # msg = _duplicates(nodes, nodesu, nnodes, name='nodes', msg=msg)
        # if lc[1]:
            # msg = _duplicates(faces, facesu, nfaces, name='faces', msg=msg)
        # raise Exception(msg)

    # # Test for unused nodes
    # facesu = np.unique(facesu)
    # c0 = np.all(facesu >= 0) and facesu.size == nnodes
    # if not c0:
        # ino = str([ii for ii in range(0, nnodes) if ii not in facesu])
        # msg = "Unused nodes in ddata[{0}]:\n".format(key)
        # msg += "    - unused nodes indices: {}".format(ino)
        # warnings.warn(msg)

    # # Check counter-clockwise orientation
    # x, y = nodes[faces, 0], nodes[faces, 1]
    # orient = ((y[:, 1] - y[:, 0])*(x[:, 2] - x[:, 1])
              # - (y[:, 2] - y[:, 1])*(x[:, 1] - x[:, 0]))

    # clock = orient > 0.
    # if np.any(clock):
        # msg = ("Some triangles not counter-clockwise\n"
               # + "  (necessary for matplotlib.tri.Triangulation)\n"
               # + "    => {}/{} triangles reshaped".format(clock.sum(), nfaces))
        # warnings.warn(msg)
        # faces[clock, 1], faces[clock, 2] = faces[clock, 2], faces[clock, 1]
    # return faces

# DEPRECATED ?
# def _check_mesh_temp(data=None, key=None):
    # # Check if provided data is mesh (as a dict)

    # # ------------
    # # Check basics
    # lmok = ['rect', 'tri', 'quadtri']
    # c0 = (
        # isinstance(data, dict)
        # and all([ss in data.keys() for ss in ['type']])
        # and data['type'] in lmok
        # and (
            # (
                # data['type'] == 'rect'
                # and all([ss in data.keys() for ss in ['R', 'Z']])
                # and isinstance(data['R'], np.ndarray)
                # and isinstance(data['Z'], np.ndarray)
                # and data['R'].ndim in [1, 2]
                # and data['Z'].ndim in [1, 2]
            # )
            # or (
                # data['type'] in ['tri', 'quadtri', 'quad']
                # and all([ss in data.keys() for ss in ['nodes', 'faces']])
                # and isinstance(data['nodes'], np.ndarray)
                # and isinstance(data['faces'], np.ndarray)
                # and data['nodes'].ndim == 2
                # and data['faces'].ndim == 2
                # and data['faces'].dtype == np.int
                # and data['nodes'].shape[1] == 2
                # and (
                    # (
                        # data['type'] in ['tri', 'quadtri']
                        # and data['faces'].shape[1] == 3
                    # )
                    # or (
                        # data['type'] == 'quad'
                        # and data['faces'].shape[1] == 4
                    # )
                # )
                # and np.max(data['faces']) <= data['nodes'].shape[0]
            # )
        # )
    # )
    # if not c0:
        # msg = (
            # """
            # A mesh should be a dict of one of the following form:

                # dict(
                 # 'type': 'rect',
                 # 'R': np.ndarray (with ndim in [1, 2]),
                 # 'Z': np.ndarray (with ndim in [1, 2]),
                 # 'shapeRZ': ('R', 'Z') or ('Z', 'R')
                # )

                 # dict(
                 # 'type': 'tri' or 'quadtri',
                 # 'nodes': np.ndarray of shape (N, 2),
                 # 'faces': np.ndarray of int of shape (N, 3)
                # )

                # dict(
                 # 'type': 'quad',
                 # 'nodes': np.ndarray of shape (N, 2),
                 # 'faces': np.ndarray of int of shape (N, 4)
                # )

            # Provided:
            # {}
            # """.format(data)
        # )
        # raise Exception(msg)

    # # ------------
    # # Check per type
    # if data['type'] == 'rect':

        # shapeRZ = data.get('shapeRZ', [None, None])
        # if shapeRZ is None:
            # shapeRZ = [None, None]
        # else:
            # shapeRZ = list(shapeRZ)

        # R, shapeRZ = _get_RZ(data['R'], name='R', shapeRZ=shapeRZ)
        # Z, shapeRZ = _get_RZ(data['Z'], name='Z', shapeRZ=shapeRZ)
        # shapeRZ = tuple(shapeRZ)

        # if shapeRZ not in [('R', 'Z'), ('Z', 'R')]:
            # msg = "Inconsistent shapeRZ"
            # raise Exception(msg)

        # def trifind(
            # r, z,
            # Rbin=0.5*(R[1:] + R[:-1]),
            # Zbin=0.5*(Z[1:] + Z[:-1]),
            # nR=R.size, nZ=Z.size,
            # shapeRZ=shapeRZ
        # ):
            # indR = np.searchsorted(Rbin, r)
            # indZ = np.searchsorted(Zbin, z)
            # indR[(r < R[0]) | (r > R[-1])] = -1
            # indZ[(z < Z[0]) | (z > Z[-1])] = -1
            # return indR, indZ
            # # if shapeRZ == ('R', 'Z'):
            # #     indpts = indR*nZ + indZ
            # # else:
            # #     indpts = indZ*nR + indR
            # # indout = ((r < R[0]) | (r > R[-1])
            # #           | (z < Z[0]) | (z > Z[-1]))
            # # indpts[indout] = -1
            # # return indpts

        # data['R'] = R
        # data['Z'] = Z
        # data['shapeRZ'] = shapeRZ
        # data['nR'] = R.size
        # data['nZ'] = Z.size
        # data['shape'] = (R.size, Z.size)
        # data['trifind'] = trifind
        # data['ftype'] = data.get('ftype', 0)

        # if data['ftype'] != 0:
            # msg = "Linear interpolation not handled yet !"
            # raise Exception(msg)

    # else:
        # # Check mesh conformity for triangulation
        # data['faces'] = _check_trimesh_conformity(
            # nodes=data['nodes'], faces=data['faces'], key=key
        # )

        # data['nnodes'] = data['nodes'].shape[0]
        # data['nfaces'] = data['faces'].shape[0]
        # data['ftype'] = data.get('ftype', 0)

        # # Convert 'quad' to 'quadtri' if relevant
        # if data['type'] == 'quad':
            # # Convert to tri mesh (solution for unstructured meshes)
            # faces = np.empty((data['nfaces']*2, 3), dtype=int)
            # faces[::2, :] = data['faces'][:, :3]
            # faces[1::2, :-1] = data['faces'][:, 2:]
            # faces[1::2, -1] = data['faces'][:, 0]
            # data['faces'] = faces
            # data['type'] = 'quadtri'
            # data['ntri'] = 2

            # # Re-check mesh conformity
            # data['faces'] = _check_trimesh_conformity(
                # nodes=data['nodes'], faces=data['faces'], key=key
            # )

        # # Check ntri
        # if data['type'] == 'tri':
            # data['ntri'] = 1
        # elif 'ntri' not in data.keys():
            # msg = (
                # """
                # For ddata[{}] of type 'quadtri', 'ntri' must be provided
                # """.format(key)
            # )
            # raise Exception(msg)

        # # Only triangular meshes so far
        # if 'tri' in data['type']:
            # if data.get('mpltri', None) is None:
                # data['mpltri'] = mplTri(
                    # data['nodes'][:, 0],
                    # data['nodes'][:, 1],
                    # data['faces']
                # )
            # if not isinstance(data['mpltri'], mplTri):
                # msg = (
                    # """
                    # ddata[{}]['mpltri'] must be a matplotlib Triangulation
                    # Provided:
                    # {}
                    # """.format(key, data['mpltri'])
                # )
            # assert data['ftype'] in [0, 1]
            # if data['ftype'] == 1:
                # data['shape'] = (data['nnodes'],)
            # else:
                # data['shape'] = (int(data['nfaces'] / data['ntri']),)

    # return data, data['shape']


# ALL USEFUL FROM HERE

# #############################################################################
# #############################################################################
#                          add data on mesh / bsplines
# #############################################################################


def add_data_meshbsplines_ref(
    ref=None,
    data=None,
    # ressources
    dmesh=None,
    dbsplines=None,
):

    if dmesh is None or dbsplines is None:
        return ref, data

    # ref is str
    if isinstance(ref, str):
        ref = [ref]

    # ref is tuple
    if isinstance(ref, (tuple, list)):

        # ref contains mesh
        rm = [(ii, rr) for ii, rr in enumerate(ref) if rr in dmesh.keys()]
        if len(rm) > 1:
            msg = (
                "ref contains references to several meshes!\n"
                f"\t- ref: {ref}\n"
                f"\t- meshes: {rm}\n"
            )
            raise Exception(msg)

        elif len(rm) == 1:
            ref = list(ref)
            kbs = [
                k0 for k0, v0 in dbsplines.items()
                if v0['mesh'] == rm[0][1]
            ]
            if len(kbs) == 1:
                ref[rm[0][0]] = kbs[0]
            elif len(kbs) > 1:
                msg = (
                    "ref contains reference to mesh with several bsplines!\n"
                    f"\t- ref: {ref}\n"
                    f"\t- mesh bsplines: {kbs}\n"
                )
                raise Exception(msg)

        # ref contains bsplines
        rbs = [(ii, rr) for ii, rr in enumerate(ref) if rr in dbsplines.keys()]
        if len(rbs) > 1:
            msg = (
                "ref contains references to several bsplines!"
                f"\t- ref: {ref}\n"
                f"\t- splines: {rbs}\n"
            )
            raise Exception(msg)

        elif len(rbs) == 1:
            ref = np.r_[
                ref[:rbs[0][0]],
                dbsplines[rbs[0][1]]['ref'],
                ref[rbs[0][0]+1:],
            ]

            # repeat data if taken from ntri > 1 
            data = _repeat_data_ntri(
                ref=ref,
                rbs1=rbs[0][1],
                refbs=dbsplines[rbs[0][1]]['ref'],
                data=data,
                # mesh
                km=dbsplines[rbs[0][1]]['mesh'],
                dmesh=dmesh,
                dbsplines=dbsplines,
            )

    return tuple(ref), data


def _repeat_data_ntri(
    ref=None,
    rbs1=None,
    refbs=None,
    data=None,
    # mesh
    km=None,
    dmesh=None,
    dbsplines=None,
):
    """ If triangular mesh with ntri > 1 => repeat data """

    c0 = (
        dmesh[km]['type'] == 'tri'
        and dmesh[km]['ntri'] > 1
    )
    if c0:
        ntri = dmesh[km]['ntri']
        indr = ref.tolist().index(refbs[0])
        nbs = dbsplines[rbs1]['shape'][0]
        ndata = data.shape[indr]
        if ndata == nbs:
            pass
        elif ndata == nbs / ntri:
            data = np.repeat(data, ntri, axis=indr)
        else:
            msg = (
                "Mismatching data shape vs multi-triangular mesh:\n"
                f"\t- data.shape[tribs] = {ndata}\n"
                f"\t- expected {nbs} / {ntri} = {nbs / ntri}\n"
            )
            raise Exception(msg)

    return data


# #############################################################################
# #############################################################################
#                           mesh generic check
# #############################################################################


def _mesh2D_check(
    coll=None,
    domain=None,
    res=None,
    R=None,
    Z=None,
    knots=None,
    cents=None,
    trifind=None,
    key=None,
):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        excluded=list(coll.dobj.get('mesh', {}).keys())
    )

    # rect of tri ?
    lc = [
        domain is not None or (R is not None and Z is not None),
        knots is not None and cents is not None,
    ]
    if all(lc) or not any(lc):
        msg = (
            "Either domain xor (R, Z) xor (knots, cents) must be provided!\n"
            "Provided:\n"
            f"\t- domain, res: {domain}, {res}\n"
            f"\t- type(R), type(Z): {type(R)}, {type(Z)}\n"
            f"\t- type(knots), type(cents): {type(knots)}, {type(cents)}\n"
        )
        raise Exception(msg)

    elif lc[0]:
        dref, ddata, dmesh = _mesh2DRect_to_dict(
            domain=domain,
            res=res,
            R=R,
            Z=Z,
            key=key,
        )

    elif lc[1]:
        dref, ddata, dmesh = _mesh2DTri_to_dict(
            knots=knots,
            cents=cents,
            trifind=trifind,
            key=key,
        )

    return dref, ddata, dmesh


def _mesh2D_polar_check(
    coll=None,
    radius=None,
    angle=None,
    radius2d=None,
    angle2d=None,
    key=None,
    # parameters
    radius_dim=None,
    radius_quant=None,
    radius_name=None,
    radius_units=None,
    angle_dim=None,
    angle_quant=None,
    angle_name=None,
):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        excluded=list(coll.dobj.get('mesh', {}).keys())
    )

    # --------------------
    # check / format input

    krk, krc = f'{key}-r-nk', f'{key}-r-nc'
    kkr, kcr = f'{key}-r-k', f'{key}-r-c'

    kak, kac = f'{key}-ang-nk', f'{key}-ang-nc'
    kka, kca = f'{key}-ang-k', f'{key}-ang-c'

    # radius data
    c0 = (
        hasattr(radius, '__iter__')
        and np.asarray(radius).ndim == 1
        and np.unique(radius).size == np.array(radius).size
        and np.allclose(np.unique(radius), radius)
    )
    if not c0:
        msg = (
            "Arg radius must be convertible to a 1d increasing array\n"
            f"\t- Provided: {radius}"
        )
        raise Exception(msg)

    # angle data
    c0 = (
        angle is None
        or (
            hasattr(angle, '__iter__')
            and np.asarray(angle).ndim == 1
            and np.unique(angle).size == np.array(angle).size
            and np.allclose(
                np.unique(np.arctan2(np.sin(angle), np.cos(angle))),
                angle,
            )
        )
    )
    if not c0:
        msg = (
            "Arg angle either\n:"
            "\t- None: radial-only polar mesh"
            "\t- convertible to a 1d increasing array\n"
            "\t\t it must be in radians\n"
            "\t\t it must be in the [-pi; pi] interval\n"
            f"\t- Provided: {angle}"
        )
        raise Exception(msg)

    # extract data
    rknot = np.unique(radius)
    rcent = 0.5*(rknot[1:] + rknot[:-1])

    # radius2d
    dradius = _check_polar_2dquant(
        coll=coll,
        quant2d=radius2d,
        quant2d_name='radius2d',
        dim=radius_dim,
        quant=radius_quant,
        name=radius_name,
        units=radius_units,
    )

    if callable(radius2d):
        keysm = None
    else:
        keysm = coll.dobj['bsplines'][coll.ddata[radius2d]['bsplines']]['mesh']

    if angle is not None:
        aknot = np.unique(np.arctan2(np.sin(angle), np.cos(angle)))
        acent = 0.5*(aknot[1:] + aknot[:-1])
        amid = 0.5*(aknot[-1] + (2.*np.pi + aknot[0]))
        amid = np.arctan2(np.sin(amid), np.cos(amid))
        if amid < acent[0]:
            acent = np.r_[amid, acent]
        else:
            acent = np.r_[acent, amid]

    # -------
    # angle2d

    if angle2d is not None:
        dangle = _check_polar_2dquant(
            coll=coll,
            quant2d=angle2d,
            quant2d_name='angle2d',
            dim=angle_dim,
            quant=angle_quant,
            name=angle_name,
            units='rad',
        )

        # check angle units = rad
        if dangle['units'] != 'rad':
            msg = (
                "Angle units must be rad\n"
                f"\t Provided: {dangle['units']}"
            )
            raise Exception(msg)

        # check angle2d is like radius2d
        c0 = (
            (callable(radius2d) and callable(angle2d))
            or coll._ddata[radius2d]['ref'] == coll._ddata[angle2d]['ref']
        )
        if not c0:
            msg = (
                "radius2d and angle2d must be of the same type, either:\n"
                "\t- both callable\n"
                "\t- both data keys with identical ref!\n"
                f"Provided:\n"
                f"\t- radius2d: {radius2d}\n"
                f"\t- angle2d: {angle2d}\n"
            )
            raise Exception(msg)

    # --------------------
    # prepare dict

    # dref
    dref = {
        krk: {'size': rknot.size},
        krc: {'size': rcent.size},
    }

    if angle is not None:
        dref.update({
            kak: {
                'size': aknot.size,
            },
            kac: {
                'size': acent.size,
            },
        })

    # ddata
    ddata = {
        kkr: {
            'data': rknot,
            'ref': krk,
            **dradius,
        },
        kcr: {
            'data': rcent,
            'ref': krc,
            **dradius,
        },
    }

    if angle is not None:
        ddata.update({
            kka: {
                'data': aknot,
                'ref': kak,
                **dangle,
            },
            kca: {
                'data': acent,
                'ref': kac,
                **dangle,
            },
        })

    # dobj
    if angle is None:
        dmesh = {
            key: {
                'type': 'polar',
                'knots': (kkr,),
                'cents': (kcr,),
                'shape-c': rcent.shape,
                'shape-k': rknot.shape,
                'radius2d': radius2d,
                'angle2d': angle2d,
                'submesh': keysm,
                'crop': False,
            },
        }
    else:
        dmesh = {
            key: {
                'type': 'polar',
                'knots': (kkr, kka),
                'cents': (kcr, kca),
                'shape-c': (rcent.size, acent.size),
                'shape-k': (rknot.size, aknot.size),
                'radius2d': radius2d,
                'angle2d': angle2d,
                'submesh': keysm,
                'crop': False,
            },
        }

    return dref, ddata, dmesh


def _check_polar_2dquant(
    quant2d=None,
    coll=None,
    quant2d_name=None,
    # parameters
    dim=None,
    quant=None,
    name=None,
    units=None,
):

    if coll.dobj.get('bsplines') is not None:
        lok = [
            k0 for k0, v0 in coll.ddata.items()
            if v0['bsplines'] in coll.dobj['bsplines'].keys()
        ]
    else:
        lok = []

    lc = [
        callable(quant2d),
        isinstance(quant2d, str) and quant2d in lok
    ]
    if not any(lc):
        msg = (
            f"Arg {quant2d_name} must be either:\n"
            f"\t- callable: {quant2d_name} = func(R, Z)\n"
            f"\t- key to existing 2d data in {lok}\n"
            f"Provided: {quant2d}\n"
        )
        raise Exception(msg)

    # quantities
    dquant = {'dim': dim, 'quant': quant, 'name': name, 'units': units}
    if isinstance(quant2d, str):
        for k0 in dquant.keys():
            if dquant[k0] is None:
                dquant[k0] = str(coll.ddata[quant2d][k0])

    return dquant


# #############################################################################
# #############################################################################
#                           Mesh2DTri
# #############################################################################


def _mesh2DTri_conformity(knots=None, cents=None, key=None):

    # ---------------------------------
    # make sure np.ndarrays of dim = 2

    knots = np.atleast_2d(knots).astype(float)
    cents = np.atleast_2d(cents).astype(int)

    # --------------
    # check shapes

    c0 = (
        knots.shape[1] == 2
        and knots.shape[0] >= 3
        and cents.shape[1] in [3, 4]
        and cents.shape[0] >= 1
        and cents.dtype == int
    )
    if not c0:
        msg = (
            "Arg knots must be of shape (nknots>=3, 2) and "
            "arg cents must be of shape (ncents>=1, 3 or 4) and dtype = int\n"
            "Provided:\n"
            f"\t- knots.shape: {knots.shape}\n"
            f"\t- cents.shape: {cents.shape}\n"
            f"\t- cents.dtype: {cents.dtype}\n"
        )
        raise Exception(msg)

    nknots = knots.shape[0]
    ncents = cents.shape[0]

    # -------------------
    # Test for duplicates

    # knots (floats => distance)
    dist = np.full((nknots, nknots), np.nan)
    ind = np.zeros(dist.shape, dtype=bool)
    for ii in range(nknots):
        dist[ii, ii+1:] = np.sqrt(
            (knots[ii+1:, 0] - knots[ii, 0])**2
            + (knots[ii+1:, 1] - knots[ii, 1])**2
        )
        ind[ii, ii+1:] = True

    ind[ind] = dist[ind] < 1.e-6
    if np.any(ind):
        iind = np.any(ind, axis=1).nonzero()[0]
        lstr = [f'\t\t- {ii}: {ind[ii, :].nonzero()[0]}' for ii in iind]
        msg = (
            f"Non-valid mesh {key}: \n"
            f"  Duplicate knots: {ind.sum()}\n"
            f"\t- knots.shape: {cents.shape}\n"
            f"\t- duplicate indices:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # cents (indices)
    centsu = np.unique(cents, axis=0)
    if centsu.shape[0] != ncents:
        msg = (
            f"Non-valid mesh {key}: \n"
            f"  Duplicate cents: {ncents - centsu.shape[0]}\n"
            f"\t- cents.shape: {cents.shape}\n"
            f"\t- unique shape: {centsu.shape}"
        )
        raise Exception(msg)

    # -------------------------------
    # Test for unused / unknown knots

    centsu = np.unique(centsu)
    c0 = np.all(centsu >= 0) and centsu.size == nknots

    # unused knots
    ino = (~np.in1d(
        range(0, nknots),
        centsu,
        assume_unique=False,
        invert=False,
    )).nonzero()[0]

    # unknown knots
    unknown = np.setdiff1d(centsu, range(nknots), assume_unique=True)

    if ino.size > 0 or unknown.size > 0:
        msg = "Knots non-conformity identified:\n"
        if ino.size > 0:
            msg += f"\t- Unused knots indices: {ino}\n"
        if unknown.size > 0:
            msg += f"\t- Unknown knots indices: {unknown}\n"
        raise Exception(msg)

    if centsu.size < nknots:
        msg = (
            f"Unused knots in {key}:\n"
            f"\t- unused knots indices: {ino}"
        )
        warnings.warn(msg)

    elif centsu.size > nknots or centsu.max() != nknots - 1:
        unknown = np.setdiff1d(centsu, range(nknots), assume_unique=True)
        msg = (
            "Unknown knots refered to in cents!\n"
            f"\t- unknown knots: {unknown}"
        )
        raise Exception(msg)

    return cents, knots


def _mesh2DTri_clockwise(knots=None, cents=None, key=None):

    x, y = knots[cents, 0], knots[cents, 1]
    orient = (
        (y[:, 1] - y[:, 0])*(x[:, 2] - x[:, 1])
        - (y[:, 2] - y[:, 1])*(x[:, 1] - x[:, 0])
    )

    clock = orient > 0.
    if np.any(clock):
        msg = (
            "Some triangles not counter-clockwise\n"
            "  (necessary for matplotlib.tri.Triangulation)\n"
            f"    => {clock.sum()}/{cents.shape[0]} triangles reshaped"
        )
        warnings.warn(msg)
        cents[clock, 1], cents[clock, 2] = cents[clock, 2], cents[clock, 1]
    return cents


def _mesh2DTri_to_dict(knots=None, cents=None, key=None, trifind=None):

    # ---------------------
    # check mesh conformity

    cents, knots = _mesh2DTri_conformity(knots=knots, cents=cents, key=key)

    # ---------------------------------------------
    # define triangular mesh and trifinder function

    # triangular mesh
    if cents.shape[1] == 3:

        # check clock-wise triangles
        cents = _mesh2DTri_clockwise(knots=knots, cents=cents, key=key)
        ntri = 1

    # Quadrangular mesh => convert to triangular
    elif cents.shape[1] == 4:

        cents2 = np.empty((cents.shape[0]*2, 3), dtype=int)
        cents2[::2, :] = cents[:, :3]
        cents2[1::2, :-1] = cents[:, 2:]
        cents2[1::2, -1] = cents[:, 0]
        cents = cents2

        # Re-check mesh conformity
        cents, knots = _mesh2DTri_conformity(knots=knots, cents=cents, key=key)
        cents = _mesh2DTri_clockwise(knots=knots, cents=cents, key=key)
        ntri = 2

    # check trifinder
    if trifind is None:
        mpltri = mplTri(knots[:, 0], knots[:, 1], cents)
        trifind = mpltri.get_trifinder()

    # ----------------------------
    # Check on trifinder function

    assert callable(trifind), "Arg trifind must be a callable!"

    try:
        out = trifind(np.r_[0.], np.r_[0])
        assert isinstance(out, np.ndarray)
    except Exception as err:
        msg = (
            "Arg trifind must return an array of indices when fed with arrays "
            "of (R, Z) coordinates!\n"
            f"\ttrifind(np.r_[0], np.r_[0.]) = {out}"
        )
        raise Exception(msg)

    # -----------------
    # Format ouput dict

    kk = f"{key}-nk"
    kc = f"{key}-nc"
    ki = f"{key}-nind"

    kcR = f"{key}-c-R"
    kcZ = f"{key}-c-Z"
    kkR = f"{key}-k-R"
    kkZ = f"{key}-k-Z"
    kii = f"{key}-ind"

    # dref
    dref = {
        kk: {
            'size': knots.shape[0],
        },
        kc: {
            'size': cents.shape[0],
        },
        ki: {
            'size': 3,
        },
    }

    # ddata
    ddata = {
        kkR: {
            'data': knots[:, 0],
            'units': 'm',
            'quant': 'R',
            'dim': 'distance',
            'ref': kk,
        },
        kkZ: {
            'data': knots[:, 1],
            'units': 'm',
            'quant': 'Z',
            'dim': 'distance',
            'ref': kk,
        },
        kcR: {
            'data': np.mean(knots[cents, 0], axis=1),
            'units': 'm',
            'quant': 'R',
            'dim': 'distance',
            'ref': kc,
        },
        kcZ: {
            'data': np.mean(knots[cents, 1], axis=1),
            'units': 'm',
            'quant': 'Z',
            'dim': 'distance',
            'ref': kc,
        },
        kii: {
            'data': cents,
            # 'units': '',
            'quant': 'indices',
            'dim': 'indices',
            'ref': (kc, ki),
        },
    }

    # dobj
    dmesh = {
        key: {
            'type': 'tri',
            'ntri': ntri,
            'cents': (kcR, kcZ),
            'knots': (kkR, kkZ),
            'ind': kii,
            # 'ref-k': (kk,),
            # 'ref-c': (kc,),
            'shape-c': (cents.shape[0],),
            'shape-k': (knots.shape[0],),
            'func_trifind': trifind,
            'crop': False,
        },
    }
    return dref, ddata, dmesh


# #############################################################################
# #############################################################################
#                           Mesh2DRect
# #############################################################################


def _mesh2DRect_X_check(
    x=None,
    res=None,
):
    """ Returns knots (x) and associated resolution

    res can be:
        - int: numbr of mesh elements desired between knots
        - float: desired average mesh element size
        - array of floats: (one for each x, desired approximate mesh size)

    """

    # ------------
    # Check inputs

    # x
    try:
        x = np.unique(np.ravel(x).astype(float))
    except Exception as err:
        msg = "x must be convertible to a sorted, flat array of floats!"
        raise Exception(msg)

    # res
    if res is None:
        res = 10

    lc = [
        isinstance(res, (int, np.int64, np.int32)) and len(x) == 2,
        isinstance(res, (float, np.floating)) and len(x) == 2,
        isinstance(res, (list, tuple, np.ndarray)) and len(x) == len(res),
    ]
    if not any(lc):
        msg = (
            "Arg res must be:\n"
            "\t- int: nb of mesh elements along x\n"
            "\t       requires len(x) = 2\n"
            "\t- float: approximate desired mesh element size along x\n"
            "\t       requires len(x) = 2\n"
            "\t- iterable: approximate desired mesh element sizes along x\n"
            "\t       requires len(x) = len(res)\n"
        )
        raise Exception(msg)

    if lc[0]:
        x_new = np.linspace(x[0], x[1], int(res)+1)
        res_new = res
        indsep = None

    elif lc[1]:
        nb = int(np.ceil((x[1]-x[0]) / res))
        x_new = np.linspace(x[0], x[1], nb+1)
        res_new = np.mean(np.diff(x))
        indsep = None

    else:

        # check conformity
        res = np.ravel(res).astype(float)
        delta = np.diff(x)
        res_sum = res[:-1] + res[1:]
        ind = res_sum > delta + 1.e-14
        if np.any(ind):
            msg = (
                "Desired resolution is not achievable for the following:\n"
                f"res_sum: {res_sum[ind]}\n"
                f"delta  : {delta[ind]}"
            )
            raise Exception(msg)

        # compute nn
        # nn = how many pairs can fit in the interval
        npairs = np.round(delta/res_sum).astype(int)
        res_sum_new = delta / npairs

        fract = res[:-1] / res_sum

        res_new = [None for ii in range(len(x)-1)]
        x_new = [None for ii in range(len(x)-1)]
        for ii in range(len(x)-1):
            res_new[ii] = (
                res_sum_new[ii]
                * np.linspace(fract[ii], 1.-fract[ii], 2*npairs[ii])
            )
            if ii == 0:
                res_add = np.concatenate(([0], np.cumsum(res_new[ii])))
            else:
                res_add = np.cumsum(res_new[ii])
            x_new[ii] = x[ii] + res_add

        indsep = np.cumsum(npairs[:-1]*2)
        res_new = np.concatenate(res_new)
        x_new = np.concatenate(x_new)

    return x_new, res_new, indsep


def _mesh2DRect_check(
    domain=None,
    res=None,
    R=None,
    Z=None,
):

    # --------------
    # check inputs

    # (domain, res) vs (R, Z)
    lc = [
        domain is not None,
        R is not None and Z is not None,
    ]
    if all(lc) or not any(lc):
        msg = (
            "Please provide (domain, res) xor (R, Z), not both:\n"
            "Provided:\n"
            f"\t- domain, res: {domain}, {res}\n"
            f"\t- R, Z: {R}, {Z}\n"
        )
        raise Exception(msg)

    if lc[0]:
        # domain
        c0 = (
            isinstance(domain, list)
            and len(domain) == 2
            and all([
                hasattr(dd, '__iter__') and len(dd) >= 2 for dd in domain
            ])
        )
        if not c0:
            msg = (
                "Arg domain must be a list of 2 iterables of len() >= 2\n"
                f"Provided: {domain}"
            )
            raise Exception(msg)

        # res
        c0 = (
            res is None
            or np.isscalar(res)
            or isinstance(res, list) and len(res) == 2
        )
        if not c0:
            msg = (
                "Arg res must be a int, float or array or a list of 2 such\n"
                f"Provided: {res}"
            )
            raise Exception(msg)

        if np.isscalar(res) or res is None:
            res = [res, res]

        # -------------
        # check R and Z

        R, resR, indR = _mesh2DRect_X_check(domain[0], res=res[0])
        Z, resZ, indZ = _mesh2DRect_X_check(domain[1], res=res[1])

    elif lc[1]:

        # R, Z check
        c0 = (
            hasattr(R, '__iter__')
            and np.asarray(R).ndim == 1
            and np.unique(R).size == np.array(R).size
            and np.allclose(np.unique(R), R)
        )
        if not c0:
            msg = "Arg R must be convertible to a 1d increasing array"
            raise Exception(msg)

        c0 = (
            hasattr(Z, '__iter__')
            and np.asarray(Z).ndim == 1
            and np.unique(Z).size == np.array(Z).size
            and np.allclose(np.unique(Z), Z)
        )
        if not c0:
            msg = "Arg Z must be convertible to a 1d increasing array"
            raise Exception(msg)

        R = np.unique(R)
        Z = np.unique(Z)

        resR = np.diff(R)
        resZ = np.diff(Z)

        if np.allclose(resR, np.mean(resR), atol=1e-12, rtol=0):
            resR = resR[0]
            indR = None
        else:
            msg = (
                "Non-uniform resolution for user-provided rectangular mesh\n"
                f"\t- unique resR: {np.unique(resR)}\n"
                f"\t- diff resR: {np.diff(np.unique(resR))}\n"
                f"\t- resR: {resR}\n"
                )
            raise NotImplementedError(msg)
            
        if np.allclose(resZ, np.mean(resZ), atol=1e-12, rtol=0):
            resZ = resZ[0]
            indZ = None
        else:
            msg = (
                "Non-uniform resolution for user-provided rectangular mesh\n"
                f"\t- unique resZ: {np.unique(resZ)}\n"
                f"\t- diff resZ: {np.diff(np.unique(resZ))}\n"
                f"\t- resZ: {resZ}\n"
                )
            raise NotImplementedError(msg)

    return R, Z, resR, resZ, indR, indZ


def _mesh2DRect_to_dict(
    domain=None,
    res=None,
    R=None,
    Z=None,
    key=None,
):

    # --------------------
    # check / format input

    kRk, kZk = f'{key}-R-nk', f'{key}-Z-nk'
    kRc, kZc = f'{key}-R-nc', f'{key}-Z-nc'
    kkR, kkZ = f"{key}-k-R", f"{key}-k-Z"
    kcR, kcZ = f"{key}-c-R", f"{key}-c-Z"

    R, Z, resR, resZ, indR, indZ = _mesh2DRect_check(
        domain=domain,
        res=res,
        R=R,
        Z=Z,
    )
    Rcent = 0.5*(R[1:] + R[:-1])
    Zcent = 0.5*(Z[1:] + Z[:-1])

    variable = not (np.isscalar(resR) and np.isscalar(resZ))

    # --------------------
    # prepare dict

    # dref
    dref = {
        kRk: {
            'size': R.size,
        },
        kZk: {
            'size': Z.size,
        },
        kRc: {
            'size': Rcent.size,
        },
        kZc: {
            'size': Zcent.size,
        },
    }

    # ddata
    ddata = {
        kkR: {
            'data': R,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'R',
            'name': 'R',
            'ref': kRk,
        },
        kkZ: {
            'data': Z,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'ref': kZk,
        },
        kcR: {
            'data': Rcent,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'R',
            'name': 'R',
            'ref': kRc,
        },
        kcZ: {
            'data': Zcent,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'ref': kZc,
        },
    }

    # dobj
    dmesh = {
        key: {
            'type': 'rect',
            'knots': (kkR, kkZ),
            'cents': (kcR, kcZ),
            # 'ref-k': (kRk, kZk),
            # 'ref-c': (kRc, kZc),
            'shape-c': (Rcent.size, Zcent.size),
            'shape-k': (R.size, Z.size),
            'variable': variable,
            'crop': False,
        },
    }
    return dref, ddata, dmesh


def _mesh2DRect_from_croppoly(crop_poly=None, domain=None):

    # ------------
    # check inputs

    c0 = hasattr(crop_poly, '__iter__') and len(crop_poly) == 2
    lc = [
        crop_poly is None,
        (
            c0
            and isinstance(crop_poly, tuple)
            and crop_poly[0].__class__.__name__ == 'Config'
            and (isinstance(crop_poly[1], str) or crop_poly[1] is None)
        )
        or crop_poly.__class__.__name__ == 'Config',
        c0
        and all([
            hasattr(cc, '__iter__') and len(cc) == len(crop_poly[0])
            for cc in crop_poly[1:]
        ])
        and np.asarray(crop_poly).ndim == 2
    ]

    if not any(lc):
        msg = (
            "Arg config must be a Config instance!"
        )
        raise Exception(msg)

    # -------------
    # Get polyand domain

    if lc[0]:
        # trivial case
        poly = None

    else:

        # -------------
        # Get poly from input

        if lc[1]:
            # (config, structure name)

            if crop_poly.__class__.__name__ == 'Config':
                config = crop_poly
                key_struct = None
            else:
                config, key_struct = crop_poly

            # key_struct if None
            if key_struct is None:
                lk, ls = zip(*[
                    (ss.Id.Name, ss.dgeom['Surf']) for ss in config.lStructIn
                ])
                key_struct = lk[np.argmin(ls)]

            # poly
            poly = config.dStruct['dObj']['Ves'][key_struct].Poly_closed

        else:

            # make sure poly is np.ndarraya and closed
            poly = np.asarray(crop_poly).astype(float)
            if not np.allclose(poly[:, 0], poly[:, -1]):
                poly = np.concatenate((poly, poly[:, 0:1]))

        # -------------
        # Get domain from poly

        if domain is None:
            domain = [
                [poly[0, :].min(), poly[0, :].max()],
                [poly[1, :].min(), poly[1, :].max()],
            ]

    return domain, poly


# #############################################################################
# #############################################################################
#                           Mesh2DRect - select
# #############################################################################


def _select_ind_check(
    ind=None,
    elements=None,
    returnas=None,
    crop=None,
    meshtype=None,
    shape2d=None,
):

    # ----------------------
    # check basic conditions

    if shape2d:
        lc = [
            ind is None,
            isinstance(ind, tuple)
            and len(ind) == 2
            and (
                all([np.isscalar(ss) for ss in ind])
                or all([
                    hasattr(ss, '__iter__')
                    and len(ss) == len(ind[0])
                    for ss in ind
                ])
                or all([isinstance(ss, np.ndarray) for ss in ind])
            ),
            (
                np.isscalar(ind)
                or (
                    hasattr(ind, '__iter__')
                    and all([np.isscalar(ss) for ss in ind])
                )
                or isinstance(ind, np.ndarray)
            )
        ]

    else:
        lc = [
            ind is None,
            np.isscalar(ind)
            or (
                hasattr(ind, '__iter__')
                and all([np.isscalar(ss) for ss in ind])
            )
            or isinstance(ind, np.ndarray)
        ]

    # check lc
    if not any(lc):
        if shape2d:
            msg = (
                "Arg ind must be either:\n"
                "\t- None\n"
                "\t- int or array of int: int indices in mixed (R, Z) index\n"
                "\t- tuple of such: int indices in (R, Z) index respectively\n"
                f"Provided: {ind}"
            )
        else:
            msg = (
                "Arg ind must be either:\n"
                "\t- None\n"
                "\t- int or array of int: int indices\n"
                "\t- array of bool: bool indices\n"
                f"Provided: {ind}"
            )
        raise Exception(msg)

    # ----------------------
    # adapt to each case

    if lc[0]:
        pass

    elif lc[1] and shape2d:
        if any([not isinstance(ss, np.ndarray) for ss in ind]):
            ind = (
                np.atleast_1d(ind[0]).astype(int),
                np.atleast_1d(ind[1]).astype(int),
            )
        lc0 = [
            [
                isinstance(ss, np.ndarray),
                np.issubdtype(ss.dtype, np.integer),
                ss.shape == ind[0].shape,
            ]
                for ss in ind
        ]
        if not all([all(cc) for cc in lc0]):
            ltype = [type(ss) for ss in ind]
            ltypes = [
                ss.dtype if isinstance(ss, np.ndarray) else False
                for ss in ind
            ]
            lshapes = [
                ss.shape if isinstance(ss, np.ndarray) else len(ss)
                for ss in ind
            ]
            msg = (
                "Arg ind must be a tuple of 2 arrays of int of same shape\n"
                f"\t- lc0: {lc0}\n"
                f"\t- types: {ltype}\n"
                f"\t- type each: {ltypes}\n"
                f"\t- shape: {lshapes}\n"
                f"\t- ind: {ind}"
            )
            raise Exception(msg)

    elif lc[1] and not shape2d:
        if not isinstance(ind, np.ndarray):
            ind = np.atleast_1d(ind).astype(int)
        c0 = (
            np.issubdtype(ind.dtype, np.integer)
            or np.issubdtype(ind.dtype, np.bool_)
        )
        if not c0:
            msg = (
                "Arg ind must be an array of bool or int\n"
                f"Provided: {ind.dtype}"
            )
            raise Exception(msg)

    else:
        if not isinstance(ind, np.ndarray):
             ind = np.atleast_1d(ind).astype(int)
        c0 = (
            np.issubdtype(ind.dtype, np.integer)
            or np.issubdtype(ind.dtype, np.bool_)
        )
        if not c0:
            msg = (
                 "Arg ind must be an array of bool or int\n"
                 f"Provided: {ind.dtype}"
            )
            raise Exception(msg)

    # elements
    elements = ds._generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    if shape2d:
        retdef = tuple
        retok = [tuple, np.ndarray, 'tuple-flat', 'array-flat', bool]
    else:
        retdef = bool
        retok = [int, bool]

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default=retdef,
        allowed=retok,
    )

    # crop
    crop = ds._generic_check._check_var(
        crop, 'crop',
        types=bool,
        default=True,
    )

    return ind, elements, returnas, crop


def _select_check(
    elements=None,
    returnas=None,
    return_ind_as=None,
    return_neighbours=None,
):

    # elements
    elements = ds._generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default='ind',
        allowed=['ind', 'data'],
    )

    # return_ind_as
    return_ind_as = ds._generic_check._check_var(
        return_ind_as, 'return_ind_as',
        types=None,
        default=int,
        allowed=[int, bool],
    )

    # return_neighbours
    return_neighbours = ds._generic_check._check_var(
        return_neighbours, 'return_neighbours',
        types=bool,
        default=True,
    )

    return elements, returnas, return_ind_as, return_neighbours,


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def _mesh2D_bsplines(key=None, lkeys=None, deg=None):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lkeys,
    )

    # deg
    deg = ds._generic_check._check_var(
        deg, 'deg',
        types=int,
        default=2,
        allowed=[0, 1, 2, 3],
    )

    # keybs
    keybs = f'{key}-bs{deg}'

    return key, keybs, deg



# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np
import scipy.integrate as scpinteg
import astropy.units as asunits


import datastock as ds


# #############################################################################
# #############################################################################
#                           Matrix - compute
# #############################################################################


def compute(
    coll=None,
    key=None,
    key_bsplines=None,
    key_diag=None,
    key_cam=None,
    # sampling
    res=None,
    mode=None,
    method=None,
    crop=None,
    # options
    brightness=None,
    # output
    store=None,
    verb=None,
):
    """ Compute the geometry matrix using:
            - a Plasma2DRect instance with a key to a bspline set
            - a cam instance with a resolution
    """

    # -----------
    # check input
    # -----------

    (
        key,
        key_bsplines, key_mesh, key_mesh0, mtype,
        key_diag, key_cam,
        method, res, mode, crop,
        brightness,
        store, verb,
    ) = _compute_check(
        coll=coll,
        key=key,
        key_bsplines=key_bsplines,
        key_diag=key_diag,
        key_cam=key_cam,
        # sampling
        method=method,
        res=res,
        mode=mode,
        crop=crop,
        # options
        brightness=brightness,
        # output
        store=store,
        verb=verb,
    )

    # -----------
    # prepare
    # -----------

    key_kR = coll.dobj['mesh'][key_mesh0]['knots'][0]
    radius_max = np.max(coll.ddata[key_kR]['data'])

    shapebs = coll.dobj['bsplines'][key_bsplines]['shape']

    # prepare indices
    indbs = coll.select_ind(
        key=key_bsplines,
        returnas=bool,
        crop=crop,
    )

    # prepare matrix
    is3d = False
    if mtype == 'polar':
        radius2d = coll.dobj[coll._which_mesh][key_mesh]['radius2d']
        r2d_reft = coll.get_time(key=radius2d)[2]
        if r2d_reft is not None:
            r2d_nt = coll.dref[r2d_reft]['size']
            if r2d_nt > 1:
                shapemat = tuple(np.r_[r2d_nt, None, indbs.sum()])
                is3d = True

    if not is3d:
        shapemat = tuple(np.r_[None, indbs.sum()])

    if verb is True:
        msg = f"Geom matrix for diag '{key_diag}':"
        print(msg)

    # -----------
    # compute
    # -----------

    if method == 'los':
        dout, units, axis = _compute_los(
            coll=coll,
            key_bsplines=key_bsplines,
            key_diag=key_diag,
            key_cam=key_cam,
            # sampling
            indbs=indbs,
            res=res,
            mode=mode,
            radius_max=radius_max,
            # groupby=groupby,
            is3d=is3d,
            # other
            shapemat=shapemat,
            brightness=brightness,
            verb=verb,
        )

    else:
        raise NotImplementedError()

    # ---------------
    # store / return
    # ---------------

    if store:
        _store(
            coll=coll,
            key=key,
            key_bsplines=key_bsplines,
            key_diag=key_diag,
            key_cam=key_cam,
            method=method,
            res=res,
            crop=crop,
            dout=dout,
            units=units,
            axis=axis,
        )

    else:
        return mat


# ###################
#   checking
# ###################


def _compute_check(
    coll=None,
    key=None,
    key_bsplines=None,
    key_diag=None,
    key_cam=None,
    # sampling
    method=None,
    res=None,
    mode=None,
    crop=None,
    # options
    brightness=None,
    # output
    store=None,
    verb=None,
):

    # key
    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('geom matrix', {}),
        short='gmat',
        key=key,
    )

    # key_bsplines
    lk = list(coll.dobj.get('bsplines', {}).keys())
    key_bsplines = ds._generic_check._check_var(
        key_bsplines, 'key_bsplines',
        types=str,
        allowed=lk,
    )

    # key_mesh0
    key_mesh = coll.dobj['bsplines'][key_bsplines]['mesh']
    mtype = coll.dobj['mesh'][key_mesh]['type']
    if mtype == 'polar':
        key_mesh0 = coll.dobj['mesh'][key_mesh]['submesh']
    else:
        key_mesh0 = key_mesh

    # key_diag, key_cam
    key_diag, key_cam = coll.get_diagnostic_cam(key=key_diag, key_cam=key_cam)

    # method
    method = ds._generic_check._check_var(
        method, 'method',
        default='los',
        types=str,
        allowed=['los'],
    )

    # res
    res = ds._generic_check._check_var(
        res, 'res',
        default=0.01,
        types=float,
        sign='> 0.',
    )

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        default='abs',
        types=str,
        allowed=['abs', 'rel'],
    )

    # crop
    crop = ds._generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )
    crop = (
        crop
        and coll.dobj['bsplines'][key_bsplines]['crop'] not in [None, False]
    )

    # brightness
    brightness = ds._generic_check._check_var(
        brightness, 'brightness',
        types=bool,
        default=False,
    )

    # store
    store = ds._generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )

    # verb
    if verb is None:
        verb = True
    if not isinstance(verb, bool):
        msg = (
            f"Arg verb must be a bool!\n"
            f"\t- provided: {verb}"
        )
        raise Exception(msg)

    return (
        key,
        key_bsplines, key_mesh, key_mesh0, mtype,
        key_diag, key_cam,
        method, res, mode, crop,
        brightness,
        store, verb,
    )


# ###################
#   compute_los                   
# ###################


def _compute_los(
    coll=None,
    key_bsplines=None,
    key_diag=None,
    key_cam=None,
    # sampling
    indbs=None,
    res=None,
    mode=None,
    key_integrand=None,
    radius_max=None,
    is3d=None,
    # other
    shapemat=None,
    brightness=None,
    verb=None,
):

    # ----------------
    # loop on cameras

    dout = {}
    doptics = coll.dobj['diagnostic'][key_diag]['doptics']
    for k0 in key_cam:

        npix = coll.dobj['camera'][k0]['dgeom']['pix_nb']
        key_los = doptics[k0]['los']

        sh = tuple([npix if ss is None else ss for ss in shapemat])
        mat = np.zeros(sh, dtype=float)

        # -----------------------
        # loop on group of pixels (to limit memory footprint)

        for ii in range(npix):

            # verb
            if verb is True:
                msg = f"\t camera '{k0}': pixel {ii + 1} / {npix}"
                end = '\n' if ii == npix - 1 else '\r'
                print(msg, flush=True, end=end)

            # sample los
            R, Z, length = coll.sample_rays(
                key=key_los,
                res=res,
                mode=mode,
                segment=None,
                ind_flat=ii,
                radius_max=radius_max,
                concatenate=False,
                return_coords=['R', 'z', 'ltot'],
            )

            # -------------
            # interpolate

            datai, units, refi = coll.interpolate_profile2d(
                key=key_bsplines,
                R=R[0],
                Z=Z[0],
                grid=False,
                azone=None,
                indbs=indbs,
                details=True,
                reshape=False,
                crop=None,
                nan0=True,
                val_out=np.nan,
                return_params=False,
                store=False,
            )

            axis = refi.index(None)
            iok = np.isfinite(datai)

            if not np.any(iok):
                continue

            datai[~iok] = 0.

            # ------------
            # integrate

            assert datai.ndim in [2, 3], datai.shape

            # integrate
            if is3d:
                mat[:, ii, :] = scpinteg.simpson(
                    datai,
                    x=length[0],
                    axis=axis,
                )
            elif datai.ndim == 3 and datai.shape[0] == 1:
                mat[ii, :] = scpinteg.simpson(
                    datai[0, ...],
                    x=length[0],
                    axis=axis,
                )
                # mat[ii, :] = np.nansum(mati[0, ...], axis=0) * reseff[ii]
            else:
                mat[ii, :] = scpinteg.simpson(
                    datai,
                    x=length[0],
                    axis=axis,
                )

        # --------------
        # post-treatment

        # brightness
        if brightness is False:
            ketend = doptics[k0]['etendue']
            etend = coll.ddata[ketend]['data']
            sh_etend = [-1 if aa == axis else 1 for aa in range(len(refi))]
            mat *= etend.reshape(sh_etend)

        # set ref
        refi = list(refi)
        refi[axis] = coll.dobj['camera'][k0]['dgeom']['ref_flat']
        refi = tuple(np.r_[refi[:axis], refi[axis], refi[axis+1:]])

        # fill dout
        dout[k0] = {
            'data': mat,
            'ref': refi,
        }

    # -----
    # units

    units = asunits.m
    if brightness is False:
        units = units * coll.ddata[ketend]['units']

    return dout, units, axis


# ###################
#   compute_vos                   
# ###################


def _compute_vos(
    coll=None,
    is2d=None,
    key_diag=None,
    key_cam=None,
    res=None,
    mode=None,
    key_integrand=None,
    radius_max=None,
    groupby=None,
    val_init=None,
    brightness=None,
):



    return None, None


# ###################
#   storing                   
# ###################


def _store(
    coll=None,
    key=None,
    key_bsplines=None,
    key_diag=None,
    key_cam=None,
    method=None,
    res=None,
    crop=None,
    dout=None,
    units=None,
    axis=None,
):

    # add data
    ddata = {}
    for k0, v0 in dout.items():
        ki = f"{key}_{k0}"
        ddata[ki] = {
            'data': v0['data'],
            'ref': v0['ref'],
            'units': units,
        }

    # shapes
    shapes = [v0['data'].shape for v0 in dout.values()]
    assert all([len(ss) == len(shapes[0]) for ss in shapes[1:]])
    shapes = np.array(shapes)
    assert np.allclose(shapes[1:, :axis], shapes[0:1, :axis])
    assert np.allclose(shapes[1:, axis+1:], shapes[0:1, axis+1:])

    # add matrix obj
    dobj = {
        'geom matrix': {
            key: {
                'data': [f"{key}_{k0}" for k0 in key_cam],
                'bsplines': key_bsplines,
                'diagnostic': key_diag,
                'camera': key_cam,
                'method': method,
                'res': res,
                'crop': crop,
                'shape': tuple(shapes[0, :]),
                'axis_chan': axis,
            },
        },
    }

    coll.update(ddata=ddata, dobj=dobj)


# ##################################################################
# ##################################################################
#               retrofit                   
# ##################################################################


def _concatenate(
    coll=None,
    key=None,
):

    # ------------
    # check inputs

    lok = list(coll.dobj.get('geom matrix', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # -----------
    # concatenate

    key_data = coll.dobj['geom matrix'][key]['data']
    key_cam = coll.dobj['geom matrix'][key]['camera']
    axis = coll.dobj['geom matrix'][key]['axis_chan']

    ref = list(coll.ddata[key_data[0]]['ref'])
    ref[axis] = None

    ind = 0
    dind = {}
    ldata = []
    for ii, k0 in enumerate(key_cam):
        datai = coll.ddata[key_data[ii]]['data']
        dind[k0] = ind + np.arange(0, datai.shape[axis])
        ldata.append(datai)
        ind += datai.shape[axis]

    data = np.concatenate(ldata, axis=axis)

    return data, ref, dind



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: initializedcheck=False
#
from cython.parallel import prange
from cpython.array cimport array, clone
from libc.math cimport sqrt as c_sqrt
from libc.math cimport NAN as CNAN
from libc.math cimport pi as c_pi
from libc.stdlib cimport malloc, free
#
cdef double _VSMALL = 1.e-9
cdef double _SMALL = 1.e-6
cdef double _TWOPI = 2.0 * c_pi


# ==============================================================================
# =  Point in path
# ==============================================================================
cdef inline bint is_point_in_path(const int nvert,
                                  const double* vertx,
                                  const double* verty,
                                  const double testx,
                                  const double testy) nogil:
    """
    Computes if a point of coordiates (testx, testy) is in the polygon defined
    by nvert vertices of coordinates (vertx, verty).

    !! -- WARNING: the poly should be CLOSED and nvert = vpoly.shape[1]-1 -- !!
    Params
    ======
        nvert : int
           number of DIFFERENT vertices in polygon
        vertx : double array
           x-coordinates of polygon
        verty : double array
           y-coordinate of polygon
        testx : double
           x-coordinate of point to be tested if in or out of polygon
        testy : double
           y-coordinate of point to be tested if in or out of polygon
    Returns
    =======
        bool : True if point is in the polygon, else False
    """
    cdef int i
    cdef bint c = 0
    for i in range(nvert):
        if ( ((verty[i]>testy) != (verty[i+1]>testy)) and
            (testx < (vertx[i+1]-vertx[i]) * (testy-verty[i])
             / (verty[i+1]-verty[i]) + vertx[i]) ):
            c = not c
    return c

cdef inline int is_point_in_path_vec(const int nvert,
                                     const double* vertx,
                                     const double* verty,
                                     const int npts,
                                     const double* testx,
                                     const double* testy,
                                     int* is_in_path) nogil:
    """
    Computes if a series of points of coordiates (testx[ii], testy[ii]) is in
    the polygon defined by nvert vertices of coordinates (vertx, verty)
    Params
    ======
        nvert : int
           number of vertices in polygon
        vertx : double array
           x-coordinates of polygon
        verty : double array
           y-coordinate of polygon
        npts : int
           number of points to test if in poly or not
        testx : double array
           x-coordinates of points to be tested if in or out of polygon
        testy : double array
           y-coordinates of points to be tested if in or out of polygon
        is_in_path : bint array
           True if point is in the polygon, else False
    Returns
    =======
     The number of "true"
    """
    cdef int ii
    cdef int tot_true = 0
    cdef bint c = 0
    for ii in range(npts):
        c = is_point_in_path(nvert, vertx, verty, testx[ii], testy[ii])
        is_in_path[ii] = c
        tot_true += c*1
    return tot_true


# ==============================================================================
# =  Computing inverse of vector and sign of each element
# ==============================================================================
cdef inline void compute_inv_and_sign(const double[3] ray_vdir,
                                      int[3] sign,
                                      double[3] inv_direction) nogil:
    """
    Computes the inverse direction and sign of each coordinate of a vector.
    Params
    ======
    ray_vdir : (3) double array
       direction of the LOS
    sign : (3) int array <INOUT>
       for each coordinate of the direction, indicates if negative or not
       If sign[i] = 1, ray_vdir[i] < 0, else sign[i] = 0
    inv_direction : (3) double array
       Inverse on each axis of direction of LOS
    """
    # computing sign and direction
    for  ii in range(3):
        if ray_vdir[ii] * ray_vdir[ii] < _VSMALL:
            inv_direction[ii] = 1. / _VSMALL
        else:
            inv_direction[ii] = 1. / ray_vdir[ii]
        if ray_vdir[ii] < 0.:
            sign[ii] = 1
        else:
            sign[ii] = 0
    return


# ==============================================================================
# =  Computing Hypotenuse
# =============================================================================
cdef inline array compute_hypot(const double[::1] xpts, const double[::1] ypts,
                                int npts=-1):
    cdef int ii
    cdef array hypot
    cdef double* ptr_hypot
    if npts == -1:
        npts  = xpts.shape[0]
    hypot = clone(array('d'), npts, False)
    ptr_hypot = hypot.data.as_doubles
    with nogil:
        for ii in range(npts):
            ptr_hypot[ii] = c_sqrt(xpts[ii] * xpts[ii] + ypts[ii] * ypts[ii])
    return hypot


cdef inline double comp_min_hypot(const double[::1] xpts,
                                  const double[::1] ypts,
                                  int npts=-1) nogil:
    cdef int ii
    cdef double tmp
    cdef double hypot = xpts[0]*xpts[0] + ypts[0]*ypts[0]
    if npts == -1:
        npts  = xpts.shape[0]
    for ii in range(npts):
        tmp = xpts[ii]*xpts[ii] + ypts[ii]*ypts[ii]
        if tmp < hypot:
            hypot = tmp
    return c_sqrt(hypot)


cdef inline double comp_min(double[::1] vec, int npts) nogil:
    cdef int ii
    cdef double res = vec[0]
    for ii in range(1,npts):
        if vec[ii] < res:
            res = vec[ii]
    return res


# ==============================================================================
# == VECTOR CALCULUS HELPERS
# ==============================================================================
cdef inline void compute_cross_prod(const double[3] vec_a,
                                    const double[3] vec_b,
                                    double[3] res) nogil:
    res[0] = vec_a[1]*vec_b[2] - vec_a[2]*vec_b[1]
    res[1] = vec_a[2]*vec_b[0] - vec_a[0]*vec_b[2]
    res[2] = vec_a[0]*vec_b[1] - vec_a[1]*vec_b[0]
    return


cdef inline double compute_dot_prod(const double[3] vec_a,
                                    const double[3] vec_b) nogil:
    return vec_a[0] * vec_b[0] + vec_a[1] * vec_b[1] + vec_a[2] * vec_b[2]


cdef inline double compute_norm(const double[3] vec) nogil:
    return c_sqrt(vec[0] * vec[0] + vec[1] * vec[1] + vec[2] * vec[2])


cdef inline double compute_g(double s, double m2b2, double rm0sqr,
                             double m0sqr, double b1sqr) nogil:
    return s + m2b2 - rm0sqr * s / c_sqrt(m0sqr * s * s + b1sqr)


cdef inline double compute_bisect(double m2b2, double rm0sqr,
                                  double m0sqr, double b1sqr,
                                  double smin, double smax) nogil:
    cdef int max_iterations = 10000
    cdef double root = 0.
    root = compute_find(m2b2, rm0sqr, m0sqr, b1sqr,
                        smin, smax, -1.0, 1.0, max_iterations, root)
    return root


cdef inline double compute_find(double m2b2, double rm0sqr,
                                double m0sqr, double b1sqr,
                                double t0, double t1, double f0, double f1,
                                int max_iterations, double root) nogil:
    cdef double fm, product

    if t0 < t1:
        # Test the endpoints to see whether F(t) is zero.
        if f0 == 0.:
            root = t0
            return root
        if f1 == 0.:
            root = t1
            return root
        if f0*f1 > 0.:
            # It is not known whether the interval bounds a root.
            return root
        for _ in range(2, max_iterations + 1):
            root = 0.5 * (t0 + t1)
            if root == t0 or root == t1:
                # The numbers t0 and t1 are consecutive floating-point
                # numbers.
                break
            fm = compute_g(root, m2b2, rm0sqr, m0sqr, b1sqr)
            product = fm * f0
            if product < 0.:
                t1 = root
                f1 = fm
            elif product > 0.:
                t0 = root
                f0 = fm
            else:
                break
        return root
    else:
        return root


# ==============================================================================
# =  Tiling
# ==============================================================================
cdef inline void tile_3_to_2d(double v0, double v1, double v2,
                              int npts, double[:,::1] res) nogil:
    """
    This function will probably not be very useful but might be used for
    inspiration (if indeed faster than using numpy in cython).
    It creates an array of shape (3, npts) as such :
        [v0, ... v0]
        [v1, ... v1]
        [v2, ... v2]
    Equivalent to :
       tab = np.tile(np.r_[v0,v1,v2], (npts,1)).T
    """
    cdef int ii
    for ii in range(npts):
        res[0,ii] = v0
        res[1,ii] = v1
        res[2,ii] = v2
    return


# ==============================================================================
# =  Polygon helpers
# ==============================================================================
cdef inline int find_ind_lowerright_corner(const double[::1] xpts,
                                           const double[::1] ypts,
                                           int npts) nogil:
    cdef int ii
    cdef int res = 0
    cdef double minx = xpts[0]
    cdef double miny = ypts[0]
    for ii in range(1,npts):
        if miny > ypts[ii]:
            minx = xpts[ii]
            miny = ypts[ii]
            res = ii
        elif miny == ypts[ii]:
            if minx < xpts[ii]:
                minx = xpts[ii]
                miny = ypts[ii]
                res = ii
    return res


# ==============================================================================
# =  Distance
# ==============================================================================
cdef inline void compute_dist_pt_vec(const double pt0, const double pt1,
                                     const double pt2, int npts,
                                     const double[:, ::1] vec,
                                     double* dist) nogil:
    """
    Compute the distance between the point P = [pt0, pt1, pt2] and each point
    Q_i, where vec = {Q_0, Q_1, ..., Q_npts-1}
    """
    cdef int ii
    for ii in range(0, npts):
        dist[ii] = c_sqrt((pt0 - vec[0, ii]) * (pt0 - vec[0, ii])
                          + (pt1 - vec[1, ii]) * (pt1 - vec[1, ii])
                          + (pt2 - vec[2, ii]) * (pt2 - vec[2, ii]))
    return


cdef inline void compute_dist_vec_vec(const int npts1, const int npts2,
                                      const double[:, ::1] vec1,
                                      const double[:, ::1] vec2,
                                      double[:, ::1] dist) nogil:
    """
    Compute the distance between each point P_i and each point
    Q_i, where vec1 = {P_0, P_1, ..., P_npts1-1} and
    vec2 = {Q_0, Q_1, ..., Q_npts2-1}
    """
    cdef int ii, jj

    for ii in range(npts1):
        for jj in range(npts2):
            dist[ii,jj] = c_sqrt((vec1[0, ii] - vec2[0, jj])
                                 * (vec1[0,ii] - vec2[0, jj])
                                 + (vec1[1,ii] - vec2[1, jj])
                                 * (vec1[1,ii] - vec2[1, jj])
                                 + (vec1[2,ii] - vec2[2, jj])
                                 * (vec1[2,ii] - vec2[2, jj]))
    return


cdef inline void compute_diff_div(const double[:, ::1] vec1,
                                  const double[:, ::1] vec2,
                                  const double* div,
                                  const int npts,
                                  double[:, ::1] res) nogil:
    """
    Computes :
      res = (vec1 - vec2) / div
    """
    cdef int ii
    cdef double invd
    for ii in range(npts):
        invd = CNAN
        if div[ii] != 0. :
            invd = 1./div[ii]
        res[0, ii] = (vec1[0,ii] - vec2[0,ii]) * invd
        res[1, ii] = (vec1[1,ii] - vec2[1,ii]) * invd
        res[2, ii] = (vec1[2,ii] - vec2[2,ii]) * invd
    return


# ==============================================================================
# == Matrix sum (np.sum)
# ==============================================================================
cdef inline void sum_by_rows(double *orig, double *out,
                             int n_rows, int n_cols) nogil:
    cdef int b, i, j
    cdef int left
    cdef int max_r = 8
    cdef int n_blocks = n_rows/max_r
    cdef double* res
    # .. initialization
    res = <double*>malloc(max_r*sizeof(double))
    for b in prange(n_blocks):
        for i in range(max_r):
            res[i] = 0
        for j in range(n_cols):
            for i in range(max_r): #calculate sum for max_r-rows simultaneously
                res[i]+=orig[(b*max_r+i)*n_cols+j]
        for i in range(max_r):
            out[b*max_r+i]=res[i]
    # left_overs:
    left = n_rows - n_blocks*max_r
    for i in prange(max_r):
        res[i] = 0
    for j in prange(n_cols):
        for i in range(left): #calculate sum for left rows simultaneously
            res[i]+=orig[(n_blocks*max_r)*n_cols+j]
    for i in prange(left):
        out[n_blocks*max_r+i]=res[i]
    free(res)
    return


cdef inline long sum_naive_int(long* orig, int n_cols) nogil:
    cdef int ii
    cdef long out

    with nogil:
        out = 0
        for ii in prange(n_cols):
            out += orig[ii]
    return out



# -*- coding: utf-8 -*-


import numpy as np
import scipy.constants as scpct
import matplotlib.colors as mcolors
import datastock as ds


from . import _utils_surface3d
from ..geom._comp_solidangles import _check_polygon_2d, _check_polygon_3d


# #############################################################################
# #############################################################################
#                       Generic for 3d surfaces
# #############################################################################


def _add_surface3d(
    coll=None,
    key=None,
    which=None,
    which_short=None,
    # 2d outline
    outline_x0=None,
    outline_x1=None,
    cent=None,
    # 3d outline
    poly_x=None,
    poly_y=None,
    poly_z=None,
    # normal vector
    nin=None,
    e0=None,
    e1=None,
    # extenthalf
    extenthalf=None,
    # curvature
    curve_r=None,
    make_planar=None,
):

    # ------------
    # check inputs

    # key
    key = ds._generic_check._obj_key(
        d0=coll.dobj.get(which, {}), short=which_short, key=key,
    )

    # geometry
    (
        cent,
        outline_x0, outline_x1,
        poly_x, poly_y, poly_z,
        nin, e0, e1,
        extenthalf, area, curve_r, gtype,
    ) = _utils_surface3d._surface3d(
        key=key,
        # 2d outline
        outline_x0=outline_x0,
        outline_x1=outline_x1,
        cent=cent,
        # 3d outline
        poly_x=poly_x,
        poly_y=poly_y,
        poly_z=poly_z,
        # normal vector at cent
        nin=nin,
        e0=e0,
        e1=e1,
        # extenthalf
        extenthalf=extenthalf,
        # curvature
        curve_r=curve_r,
        make_planar=make_planar,
    )

    # ----------
    # create dict

    # keys
    knpts = f'{key}-npts'
    if gtype != '3d':
        kp0 = f'{key}-x0'
        kp1 = f'{key}-x1'
        outline = (kp0, kp1)
        poly = None
        npts = outline_x0.size
    else:
        kpx = f'{key}-x'
        kpy = f'{key}-y'
        kpz = f'{key}-z'
        poly = (kpx, kpy, kpz)
        outline = None
        npts = poly_x.size

    # refs
    dref = {
        knpts: {'size': npts},
    }

    # data
    if gtype != '3d':
        ddata = {
            kp0: {
                'data': outline_x0,
                'ref': knpts,
                'dim': 'distance',
                'name': 'x0',
                'quant': 'x0',
                'units': 'm',
            },
            kp1: {
                'data': outline_x1,
                'ref': knpts,
                'dim': 'distance',
                'name': 'x1',
                'quant': 'x1',
                'units': 'm',
            },
        }
    else:
        ddata = {
            kpx: {
                'data': poly_x,
                'ref': knpts,
                'dim': 'distance',
                'name': 'x',
                'quant': 'x',
                'units': 'm',
            },
            kpy: {
                'data': poly_y,
                'ref': knpts,
                'dim': 'distance',
                'name': 'y',
                'quant': 'y',
                'units': 'm',
            },
            kpz: {
                'data': poly_z,
                'ref': knpts,
                'dim': 'distance',
                'name': 'z',
                'quant': 'z',
                'units': 'm',
            },
        }

    # dobj
    dobj = {
        which: {
            key: {
                'dgeom': {
                    'type': gtype,
                    'curve_r': curve_r,
                    'outline': outline,
                    'extenthalf': extenthalf,
                    'poly': poly,
                    'area': area,
                    'cent': cent,
                    'nin': nin,
                    'e0': e0,
                    'e1': e1,
                },
            },
        },
    }

    return dref, ddata, dobj


# #############################################################################
# #############################################################################
#                       Generic dmisc for 3d surfaces
# #############################################################################


def _dmisc(key=None, color=None):

    # --------
    # color

    if color is None:
        color = 'k'
        if not mcolors.is_color_like(color):
            msg = (
                f"Arg color for '{key}' must be a matplotlib color!\n"
                f"Provided: {color}\n"
            )
            raise Exception(msg)

    color = mcolors.to_rgba(color)

    return {'color': color}


# #############################################################################
# #############################################################################
#                       Utilities
# #############################################################################


def _return_as_dict(
    coll=None,
    key=None,
):
    """ Return as dict the following objects

    - camera
    - aperture
    - filter (treated as aperture)

    useful input for low-level routines

    """

    if isinstance(key, str):
        key = [key]

    lcam = list(coll.dobj.get('camera', {}).keys())
    lap = list(coll.dobj.get('aperture', {}).keys())
    lfilt = list(coll.dobj.get('filter', {}).keys())
    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=(list, tuple),
        types_iter=str,
        allowed=lcam + lap + lfilt,
    )

    dout = {}
    for k0 in key:

        if k0 in lcam:
            cls = 'camera'
        elif k0 in lap:
            cls = 'aperture'
        elif k0 in lfilt:
            cls = 'filter'

        # compute
        if cls == 'camera':

            dout[k0] = coll.get_camera_unit_vectors(key=k0)
            dgeom = coll.dobj['camera'][k0]['dgeom']
            cx, cy, cz = coll.get_camera_cents_xyz(key=k0)

            dout[k0].update({
                'outline_x0': coll.ddata[dgeom['outline'][0]]['data'],
                'outline_x1': coll.ddata[dgeom['outline'][1]]['data'],
                'cents_x': cx,
                'cents_y': cy,
                'cents_z': cz,
                'pix_area': dgeom['pix_area'],
                'parallel': dgeom['parallel'],
            })

        elif cls in ['aperture', 'filter']:

            ap = coll.dobj[cls][k0]['dgeom']
            dout[k0] = {
                'cent': ap['cent'],
                'poly_x': coll.ddata[ap['poly'][0]]['data'],
                'poly_y': coll.ddata[ap['poly'][1]]['data'],
                'poly_z': coll.ddata[ap['poly'][2]]['data'],
                'nin': ap['nin'],
                'e0': ap.get('e0'),
                'e1': ap.get('e1'),
            }

    return dout



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from ._class06_Grating import Grating as Previous
from . import _class3_check
from . import _class7_check as _check
from . import _class7_compute as _compute


__all__ = ['Camera']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class Camera(Previous):

    # _ddef = copy.deepcopy(ds.DataStock._ddef)
    # _ddef['params']['ddata'].update({
    #       'bsplines': (str, ''),
    # })
    # _ddef['params']['dobj'] = None
    # _ddef['params']['dref'] = None

    # _show_in_summary_core = ['shape', 'ref', 'group']
    _show_in_summary = 'all'

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'camera': [
            'dgeom.type',
            'dgeom.parallel',
            'dgeom.shape',
            'dgeom.ref',
            'dgeom.pix_area',
            'dgeom.pix_nb',
            'dgeom.outline',
            'dgeom.cent',
            'dgeom.cents',
            'dmat.qeff_energy',
            'dmat.qeff',
            # 'dmisc.color',
        ],
    })

    def _add_camera(
        self,
        dref=None,
        ddata=None,
        dobj=None,
        dmat=None,
        color=None,
    ):
        key = list(dobj['camera'].keys())[0]

        # material
        dref2, ddata2, dmat = _check._dmat(
            key=key,
            dmat=dmat,
        )

        if dmat is not None:
            if dref2 is not None:
                dref.update(dref2)
                ddata.update(ddata2)
            dobj['camera'][key]['dmat'] = dmat

        # dmisc
        dobj['camera'][key]['dmisc'] = _class3_check._dmisc(
            key=key,
            color=color,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

    def add_camera_1d(
        self,
        key=None,
        # geometry
        dgeom=None,
        # quantum efficiency
        dmat=None,
        # dmisc
        color=None,
    ):
        """ add a 1d camera

        A 1d camera is an unordered set of pixels of indentical outline
        Its geometry os defined by dgeom
        Its material properties (i.e: quantum efficiency) in dmat

        The geometry in dgeom must contain:
            - 'outline_x0': 1st coordinate of planar outline of a single pixel
            - 'outline_x1': 1st coordinate of planar outline of a single pixel
            - 'cents_x': x coordinate of the centers of ll pixels
            - 'cents_y': y coordinate of the centers of ll pixels
            - 'cents_z': z coordinate of the centers of ll pixels
            - 'nin_x': x coordinate of inward normal unit vector of all pixels
            - 'nin_y': y coordinate of inward normal unit vector of all pixels
            - 'nin_z': z coordinate of inward normal unit vector of all pixels
            - 'e0_x': x coordinate of e0 unit vector of all pixels
            - 'e0_y': y coordinate of e0 unit vector of all pixels
            - 'e0_z': z coordinate of e0 unit vector of all pixels
            - 'e1_x': x coordinate of e1 unit vector of all pixels
            - 'e1_y': y coordinate of e1 unit vector of all pixels
            - 'e1_z': z coordinate of e1 unit vector of all pixels

        The material dict, dmat can contain:
            - 'energy': a 1d energy vector , in eV
            - 'qeff': a 1d vector, same size as energy, with values in [0; 1]

        """
        # check / format input
        dref, ddata, dobj = _check._camera_1d(
            coll=self,
            key=key,
            **dgeom,
        )

        # add generic parts
        self._add_camera(
            dref=dref,
            ddata=ddata,
            dobj=dobj,
            dmat=dmat,
            color=color,
        )

    def add_camera_2d(
        self,
        key=None,
        # geometry
        dgeom=None,
        # material
        dmat=None,
        # dmisc
        color=None,
    ):
        """ add a 2d camera

        A 2d camera is an ordered 2d grid of pixels of indentical outline
        Its geometry os defined by dgeom
        Its material properties (i.e: quantum efficiency) in dmat

        The geometry in dgeom must contain:
            - 'outline_x0': 1st coordinate of planar outline of a single pixel
            - 'outline_x1': 1st coordinate of planar outline of a single pixel
            - 'cent': (x, y, z) coordinate of the center of the camera
            - 'cents_x0': x0 coordinate of the centers of all pixels
            - 'cents_x1': x1 coordinate of the centers of all pixels
            - 'nin': x coordinate of inward normal unit vector of all pixels
            - 'e0': x coordinate of e0 unit vector of all pixels
            - 'e1': x coordinate of e1 unit vector of all pixels

        The material dict, dmat can contain:
            - 'energy': a 1d energy vector , in eV
            - 'qeff': a 1d vector, same size as energy, with values in [0; 1]

        """
        # check / format input
        dref, ddata, dobj = _check._camera_2d(
            coll=self,
            key=key,
            **dgeom,
        )

        # add generic parts
        self._add_camera(
            dref=dref,
            ddata=ddata,
            dobj=dobj,
            dmat=dmat,
            color=color,
        )

    # ---------------
    # add pinhole cameras
    # ---------------

    def add_camera_pinhole(
        self,
        key=None,
        key_pinhole=None,
        key_diag=None,
        cam_type=None,
        # position
        x=None,
        y=None,
        R=None,
        z=None,
        phi=None,
        # orientation
        theta=None,
        dphi=None,
        tilt=None,
        # camera
        focal=None,
        pix_nb=None,
        pix_size=None,
        pix_spacing=None,
        # pinhole
        pinhole_radius=None,
        pinhole_size=None,
        # reflections
        reflections_nb=None,
        reflections_type=None,
        # diagnostic
        compute=None,
        config=None,
        length=None,
    ):

        return _compute.add_camera_pinhole(
            coll=self,
            key=key,
            key_pinhole=key_pinhole,
            key_diag=key_diag,
            cam_type=cam_type,
            # position
            x=x,
            y=y,
            R=R,
            z=z,
            phi=phi,
            # orientation
            theta=theta,
            dphi=dphi,
            tilt=tilt,
            # camera
            focal=focal,
            pix_nb=pix_nb,
            pix_size=pix_size,
            pix_spacing=pix_spacing,
            # pinhole
            pinhole_radius=pinhole_radius,
            pinhole_size=pinhole_size,
            # reflections
            reflections_nb=reflections_nb,
            reflections_type=reflections_type,
            # diagnostic
            compute=compute,
            config=config,
            length=length,
        )

    # ---------------
    # utilities
    # ---------------

    def get_camera_unit_vectors(self, key=None):
        """ Return unit vectors components as dict """
        return _check.get_camera_unitvectors(
            coll=self,
            key=key,
        )

    def get_camera_cents_xyz(self, key=None):
        """ Return cents_x, cents_y, cents_z """
        return _check.get_camera_cents_xyz(
            coll=self,
            key=key,
        )

    def get_as_dict(self, key=None):
        """ Return the desired object as a dict (input to some routines) """

        return _class3_check._return_as_dict(
            coll=self,
            key=key,
        )
